from fractions import *

class FFE:
    """Finite field element e, modulo a prime p"""
    def __init__(self, p, e):
        self.p = p
        self.e = e % p
    
    def identity (self):
        return FFE(self.p, 1)
    
    def __add__(self, other):
        if (self.p != other.p):
            raise ArithmeticError("Incompatible moduli")
        s = (self.e + other.e) % self.p
        return FFE(self.p, s)

    def __sub__(self, other):
        if (self.p != other.p):
            raise ArithmeticError("Incompatible moduli")
        s = (self.e - other.e) % self.p
        return FFE(self.p, s)

    def __neg__(self):
        return FFE(self.p, -self.e)
    
    def __mul__(self, other):
        if (isinstance(other, int)):
            return FFE(self.p, self.e*other)
        
        if (self.p != other.p):
            raise ArithmeticError("Incompatible moduli")
        s = (self.e * other.e) % self.p
        return FFE(self.p, s)

    def __truediv__(self, other):
        if (self.p != other.p):
            raise ArithmeticError("Incompatible moduli")
        s = (self.e * inverse(other.e, self.p)) % self.p
        return FFE(self.p, s)

    def __pow__(self, other):
        return fast_power_obj(self, other)

    def __str__(self):
        return '%d' % (self.e)

    def __repr__(self):
        return str(self)
   
    def __eq__(self, other):
        return ((self.p == other.p) and (self.e == other.e))

    def __ne__(self, other):
        return not (self == other)

## Examples
# a = FFE(13,3)
# b = FFE(13,9)
## Returns FFE(13,12), which represents 12 modulo 13 (or -1 modulo 13)
# a + b
## Verify that 12 mod 13 and -1 mod 13 are the same
# a + b == FFE(13,-1)    # Returns True
# a + b == FFE(13, 1)    # Returns False
## Verify that 3 and 9 are inverses modulo 13:
# a*b
# a*b == FFE(13,1)       # Returns True

class PolyRing:
	def __init__ (self, one, coeff):
		self.coeff = coeff
		self.one = one
		self.zero = one - one
		
		deg = len(coeff) - 1
		if (deg >= 0) and (coeff[deg] == self.zero):
			raise ArithmeticError("Leading coefficient in a polynomial cannot be zero")

	def identity (self):
		return PolyRing(self.one, [self.one])
	
	def deg (self):
		return len(self.coeff) - 1
	
	# The analogue of absolute value for a polynomial
	# is to make it monic (i.e. leading coefficient one)
	def __abs__ (self):
		deg = len(self.coeff) - 1
		topc = self.coeff[deg]
		monic_coeffs = [ x / topc for x in self.coeff ]
		return PolyRing(self.one, monic_coeffs)
				
	def __add__ (self, other):
		my_deg = len(self.coeff) - 1
		other_deg = len(other.coeff) - 1
		deg = max(my_deg, other_deg)
		
		last_nonzero_idx = -1
		sum_lst = []
		for k in range(deg+1):
			s = self.zero
			if (k <= my_deg):
				s = s + self.coeff[k]
			if (k <= other_deg):
				s = s + other.coeff[k]
			
			sum_lst.append(s)
			if (s != self.zero):
				last_nonzero_idx = k
			
		return PolyRing(self.one, sum_lst[:last_nonzero_idx+1])

	def __neg__ (self):
		minus_coeff = [ -x for x in self.coeff ]
		return PolyRing(self.one, minus_coeff)
		
	def __sub__ (self, other):
		return self + (-other)
		
	def __mul__(self, other):
		my_deg = len(self.coeff) - 1
		other_deg = len(other.coeff) - 1
		deg = my_deg + other_deg
		
		if (my_deg < 0) or (other_deg < 0):
			## one of the factors is zero, so product is zero
			zeropoly = PolyRing(self.one, [])
			return zeropoly
		
		prod_lst = []
		for n in range(deg+1):
			conv = self.zero
			for k in range(n+1):
				if (k <= my_deg) and (n - k <= other_deg):
					conv += self.coeff[k] * other.coeff[n-k]
			prod_lst.append(conv)
		
		# print(("about to make product: (%s) * (%s)" % (str(self),str(other))) + str(prod_lst))
		
		return PolyRing(self.one, prod_lst)
		
	def __str__(self):
		deg = len(self.coeff)
		
		if (deg == 0):
			return "0"

		mystr = str(self.coeff[0]) 
		
		if (deg == 1):
			return mystr
		
		for k in range(1, deg):
			mystr = mystr + " + " + str(self.coeff[k]) + "x^{" + str(k) + "}"

		return mystr
	
	def division (a, b):
		adeg = len(a.coeff) - 1
		bdeg = len(b.coeff) - 1

		if (bdeg < 0):
			raise ArithmeticError("Cannot divide by the zero polynomial")
		
		if (adeg < bdeg):
			zeropoly = PolyRing(b.one, [])
			return (zeropoly, a)

		#print("A : " + str(a))
		#print("B : " + str(b))
				
		topcoeff = [b.zero] * (adeg - bdeg)
		topcoeff.append(a.coeff[adeg] / b.coeff[bdeg])
		top = PolyRing(b.one, topcoeff)
		
		#print("A : " + str(a))
		#print("A after " + str(a - top*b))
		
		(q,r) = PolyRing.division(a - top*b, b)
		return (top + q, r)


	def __floordiv__ (self, other):
		return PolyRing.division(self,other)[0]
		
	def __mod__ (self, other):
		return PolyRing.division(self,other)[1]
		
	def __eq__ (self, other):
		my_deg = len(self.coeff) - 1
		other_deg = len(other.coeff) - 1
		
		if (my_deg != other_deg):
			return False
		
		for k in range(my_deg + 1):
			if (self.coeff[k] != other.coeff[k]):
				return False
		
		return True
		
	def __repr__(self):
		return str(self)
		
	def __pow__(self, k):
		return fast_power_obj(self, k, self.identity())

### Examples
# Make the polynomial 7x^2 + 5 in F_{19}[x]:
# a = PolyRing(FFE(19,1), [FFE(19, 5), FFE(19, 0), FFE(19,7)] )

## Make the polynomial (7/3)x^2 + 5 in ring Q[x], where Q is the rational field:
# from fractions import *
# a = PolyRing(Fraction(1), [ Fraction(5), Fraction(0), Fraction(7,3) ] )


class PolyRingModulo (PolyRing):
	def __init__ (self, p, lst):
		one = FFE(p,1)
		coeff = [ FFE(p, x) for x in lst ]
		super().__init__(one, coeff)

	def cyclic (p, n):
		return PolyRingModulo(p, [-1] + ([0]*(n-1)) + [1])

### Examples
# Make the polynomial 7x^2 + 5 in F_{19}[x]:
# a = PolyRingModulo(19, [5, 0, 7])

class PolyRingQ (PolyRing):
	"""Polynomial Ring over the Rationals"""
	def __init__ (self, lst):
		one = Fraction(1)
		coeff = []
		for x in lst:
			if type(x) == int:
				a = Fraction(x)
			else:
				a = Fraction(x[0], x[1])
			coeff.append(a)
			
		super().__init__(one, coeff)

	def cyclic (n):
		return PolyRingQ([-1] + ([0]*(n-1)) + [1])



# Note: an iterative implementation of the extended euclidean algorithm 
# uses less space and is more efficient, but the recursive algorithm
# is simpler to code and is easier to understand.
def xgcd (a, b, one=1):
	"""Extended Euclidean Algorithm."""	
	zero = one - one
	
	if (b == zero):
		# Base case:  gcd(a,0) = a and a = (1)*a + (0)*b
		# so return (d,u,v) with d=a, u=1, v=0
		return (a, one, zero)
	else:
		r = a % b
		q = (a - r) // b
		(d, u, v) = xgcd(b, r, one)
			
	return (d, v, u - q*v)

def inverse (a, m, one=1):
	(d, u, v) = xgcd(a,m,one)
	
	if (abs(d) == one):
		return (u // d) % m
	else:
		return None
		
# Note: an iterative implementation of the fast power algorithm 
# uses less space and is more efficient, but the recursive algorithm
# is simpler to code and is easier to understand.
def fast_power_obj (a, k, one=1):
    """for object a and integer k, computes a^k"""
    if (k == 0):
        return one
        
    if ((k % 2) == 1):
        # k is odd
        # recursively compute y = a^{k-1} (mod m)
        y = fast_power_obj(a, k-1, one)
	
        return (a*y)
    else:
        # k is even
        # recursively compute y = a^{k/2} (mod m)
        y = fast_power_obj(a, k//2, one)

        return (y*y)

# computes a^e modulo m.  Here, a and m are objects in a ring
# and e is an integer.
# efficient iterative algorithm
def modexp (a,e,m, one=1):
	result = one
	a_to_currpow = a 
	
	while (e != 0):
		if (e % 2) == 1:
			#exponent is odd
			result = (result * a_to_currpow) % m
			e = e - 1

		e = e // 2
		a_to_currpow = (a_to_currpow * a_to_currpow) % m		
		
	return result

### Example from class, Wed Apr 20
## Make the polynomial a = 3x^4 + 5x + 5 in F_{13}[x]
# a = PolyRingModulo(13,[5,5,0,0,3])
## Make the polynomial b = 3x^4 + 2x^3 + 5x^2 + 9x + 4 in F_{13}[x]
# b=PolyRingModulo(13,[4,9,5,2,3])

## compute d=gcd(a,b) and polynomials u,v such that d=ua + vb
# (d,u,v) = xgcd(a,b,one=b.identity())

#p=281
#a=PolyRingModulo(p,[-14,6,-1,0,0,1])
#b=PolyRingModulo(p,[-1,1,9,1])

# center lift a Finite Field Element e to the integers Z
def liftFFE (a):
	p = a.p
	e = a.e
	
	e = e % p
	if (e > (p-1)//2):
		e = e - p
	
	return e

# given a PolyRing f over finite field elements, lift to a polynomial over Q
def center_lift (f):
	coeff = [ Fraction(liftFFE(a)) for a in f.coeff ]
	return PolyRing(Fraction(1), coeff)
	
# given a PolyRing f over Q, reduce modulo p
def reduce_modulo (f, p):
	coeff = []
	for x in f.coeff:
		num = x.numerator
		denom = x.denominator
		a = FFE(p, num) / FFE(p, denom)
		coeff.append(a)
		
	return PolyRing(FFE(p,1), coeff)

# Establish the environment for HW13, #4
(N,p,q) = (7,37,479)
# Quotient ring modulus x^N - 1
mod = PolyRingQ.cyclic(N)
modq = reduce_modulo(mod,q)
modp = reduce_modulo(mod,p)

## part a
# f
f=PolyRingQ([1,-1,1,0,-1,1])
fq = reduce_modulo(f,q)
fp = reduce_modulo(f,p)

# g
g = PolyRingQ([0,-1,-1,0,1,0,1])
gq = reduce_modulo(g,q)
gp = reduce_modulo(g,p)

## part b:
## m is the message
## r is the random element
m = PolyRingQ([1,1,7,0,0,-18,4])
mq = reduce_modulo(m,q)
mp = reduce_modulo(m,p)

r = PolyRingQ([-1,-1,0,1,0,0,1])
rq = reduce_modulo(r,q)
rp = reduce_modulo(r,p)

## part c
c=PolyRingModulo(q,[197,464,276,221,415,4,350])

